"""
Term import.
"""

import csv

from lute.db import db
from lute.models.term import Status
from lute.models.language import Language
from lute.term.model import Term, Repository


class BadImportFileError(Exception):
    """
    Raised if the import file is bad:

    - unknown language or status
    - dup term
    etc.
    """


def import_file(filename, create_terms=True, update_terms=True):
    """
    Validate and import file.

    Throws BadImportFileError if file contains invalid data.
    """
    import_data = _load_import_file(filename)
    _validate_data(import_data)
    return _do_import(import_data, create_terms, update_terms)


def _load_import_file(filename, encoding="utf-8-sig"):
    "Create array of hashes from file."
    importdata = []
    with open(filename, "r", encoding=encoding) as f:
        reader = csv.DictReader(f)

        fieldnames = reader.fieldnames
        if fieldnames is None:
            raise BadImportFileError("No terms in file")
        _validate_data_fields(fieldnames)

        for line in reader:
            importdata.append(line)

    if len(importdata) == 0:
        raise BadImportFileError("No terms in file")

    return importdata


def _validate_data_fields(field_list):
    "Check the keys in the file."
    required = ["language", "term"]
    for k in required:
        if k not in field_list:
            raise BadImportFileError(f"Missing required field '{k}'")

    allowed = required + [
        "translation",
        "parent",
        "status",
        "tags",
        "pronunciation",
        "link_status",
    ]
    for k in field_list:
        if k not in allowed:
            raise BadImportFileError(f"Unknown field '{k}'")


def _validate_data(import_data):
    """
    Check the data.
    """
    _validate_languages(import_data)
    _validate_terms_exist(import_data)
    _validate_statuses(import_data)
    _validate_no_duplicate_terms(import_data)


def _create_langs_dict(import_data):
    "Create dictionary of language name to Language."
    lang_dict = {}
    langs = [hsh["language"].strip() for hsh in import_data]
    for lang_name in list(set(langs)):
        lang_dict[lang_name] = Language.find_by_name(lang_name)
    return lang_dict


def _get_status(s):
    "Convert status to db value."
    status_map = {
        "": 1,
        "1": 1,
        "2": 2,
        "3": 3,
        "4": 4,
        "5": 5,
        "W": Status.WELLKNOWN,
        "I": Status.IGNORED,
    }
    return status_map.get(s)


def _validate_languages(import_data):
    "Validate the languages."
    lang_dict = _create_langs_dict(import_data)
    for lang_name, lang in lang_dict.items():
        if lang is None:
            raise BadImportFileError(f"Unknown language '{lang_name}'")


def _validate_statuses(import_data):
    "All statuses must be valid."
    statuses = [hsh["status"].strip() for hsh in import_data if "status" in hsh]
    for s in set(statuses):
        if _get_status(s) is None:
            raise BadImportFileError(
                "Status must be one of 1, 2, 3, 4, 5, I, W, or blank"
            )


def _validate_terms_exist(import_data):
    "All records must have a term."
    blanks = [hsh for hsh in import_data if hsh["term"].strip() == ""]
    if len(blanks) > 0:
        raise BadImportFileError("Term is required")


def _validate_no_duplicate_terms(import_data):
    """
    Duplicate terms aren't allowed.

    If file contained two duplicate terms, which is the "correct" one?
    """

    def make_lang_term_string(hsh):
        t = hsh["term"].strip()
        # Have to also clear unicode whitespace.
        t = " ".join(t.split())
        return f"{hsh['language']}: {t.lower()}"

    lang_terms = [make_lang_term_string(hsh) for hsh in import_data]
    term_counts = {}
    for term in lang_terms:
        term_counts[term] = term_counts.get(term, 0) + 1
    duplicates = [term for term, count in term_counts.items() if count > 1]
    if len(duplicates) != 0:
        raise BadImportFileError(f"Duplicate terms in import: {', '.join(duplicates)}")


def _import_term_skip_parents(repo, rec, lang):
    "Add a single record to the repo."
    t = Term()
    t.language = lang
    t.language_id = lang.id
    t.text = rec["term"]
    if "translation" in rec:
        t.translation = rec["translation"]
    if "status" in rec:
        status = _get_status(rec["status"])
        if status is not None:
            t.status = int(status)
    if "pronunciation" in rec:
        t.romanization = rec["pronunciation"]
    if "tags" in rec:
        tags = list(map(str.strip, rec["tags"].split(",")))
        t.term_tags = [t for t in tags if t != ""]
    repo.add(t)


def _update_term_skip_parents(t, repo, rec):
    "Update a term in the repo."
    # Don't change the lang or text of the term
    # t.language = lang
    # t.language_id = lang.id
    # t.text = rec["term"]
    if "translation" in rec:
        t.translation = rec["translation"]
    if "status" in rec:
        status = _get_status(rec["status"])
        if status is not None:
            t.status = int(status)
    if "pronunciation" in rec:
        t.romanization = rec["pronunciation"]
    if "tags" in rec:
        tags = list(map(str.strip, rec["tags"].split(",")))
        t.term_tags = [t for t in tags if t != ""]
    repo.add(t)


def _set_term_parents(repo, rec, lang):
    "Set the term parents."
    t = repo.find(lang.id, rec["term"])
    parents = list(map(str.strip, rec["parent"].split(",")))
    t.parents = [p for p in parents if p != ""]
    if "link_status" in rec and len(parents) == 1:
        sync_status = rec["link_status"] or ""
        t.sync_status = sync_status.strip().lower() == "y"
    if len(parents) != 1:
        t.sync_status = False
    repo.add(t)


def _do_import(import_data, create_terms=True, update_terms=True):
    """
    Import records.

    If create_terms is True, create new terms.
    If update_terms is True, update existing terms.

    The import is done in two passes:
    1. import the basic terms, without setting their parents
    2. update the terms with parents

    The two passes are done because the import file may
    contain a parent in its own row, and we want that to be
    imported first to get its own specified data.
    """
    # pylint: disable=too-many-locals

    repo = Repository(db)

    skipped = 0

    # Keep track of the created and updated terms: we only want to
    # update these ones in pass #2.
    created_terms = []
    updated_terms = []

    def term_string(lang, term):
        return f"{lang.id}-{term}"

    for batch in [import_data[i : i + 100] for i in range(0, len(import_data), 100)]:
        langs_dict = _create_langs_dict(batch)
        for hsh in batch:
            lang = langs_dict[hsh["language"]]
            t = repo.find(lang.id, hsh["term"])
            ts = term_string(lang, hsh["term"])
            if t is None and create_terms:
                _import_term_skip_parents(repo, hsh, lang)
                created_terms.append(ts)
            elif t is not None and update_terms:
                _update_term_skip_parents(t, repo, hsh)
                updated_terms.append(ts)
            else:
                skipped += 1
        repo.commit()

    pass_2 = [t for t in import_data if "parent" in t and t["parent"] != ""]
    for batch in [pass_2[i : i + 100] for i in range(0, len(pass_2), 100)]:
        langs_dict = _create_langs_dict(batch)
        for hsh in batch:
            lang = langs_dict[hsh["language"]]
            ts = term_string(lang, hsh["term"])
            if ts in created_terms or ts in updated_terms:
                _set_term_parents(repo, hsh, lang)
        repo.commit()

    stats = {
        "created": len(created_terms),
        "updated": len(updated_terms),
        "skipped": skipped,
    }

    return stats



"""
Reading rendering helpers.
"""

import re
from sqlalchemy import text as sqltext

from lute.models.term import Term
from lute.parse.base import ParsedToken
from lute.read.render.renderable_calculator import RenderableCalculator
from lute.db import db


def find_all_Terms_in_string(s, language):  # pylint: disable=too-many-locals
    """
    Find all terms contained in the string s.

    For example
    - given s = "Here is a cat"
    - given terms in the db: [ "cat", "a cat", "dog" ]

    This would return the terms "cat" and "a cat".

    The code first queries for exact single-token matches,
    and then multiword matches, because that's much faster
    than querying for everthing at once.  (This may no longer
    be true, can change it later.)
    """

    # Extract word tokens from the input string
    cleaned = re.sub(r"\s+", " ", s)
    tokens = language.get_parsed_tokens(cleaned)

    parser = language.parser

    # fyi - Manually searching for terms was slow (i.e., querying for
    # all terms, and checking if the strings were in the string s).

    # Query for terms with a single token that match the unique word tokens
    word_tokens = filter(lambda t: t.is_word, tokens)
    tok_strings = [parser.get_lowercase(t.token) for t in word_tokens]
    tok_strings = list(set(tok_strings))
    terms_matching_tokens = (
        db.session.query(Term)
        .filter(
            Term.language == language,
            Term.text_lc.in_(tok_strings),
            Term.token_count == 1,
        )
        .all()
    )

    # Multiword terms have zws between all tokens.
    # Create content string with zws between all tokens for the match.
    zws = "\u200B"  # zero-width space
    lctokens = [parser.get_lowercase(t.token) for t in tokens]
    content = zws + zws.join(lctokens) + zws

    sql = sqltext(
        """
        SELECT WoID FROM words
        WHERE WoLgID=:language_id and WoTokenCount>1
        AND :content LIKE '%' || WoTextLC || '%'
        """
    )
    sql = sql.bindparams(language_id=language.id, content=content)
    idlist = db.session.execute(sql).all()
    woids = [int(p[0]) for p in idlist]
    contained_terms = db.session.query(Term).filter(Term.id.in_(woids)).all()

    # Note that the above method (querying for ids, then getting terms)
    # is faster than using the model as shown below!
    ### contained_term_query = db.session.query(Term).filter(
    ###     Term.language == language,
    ###     Term.token_count > 1,
    ###     func.instr(content, Term.text_lc) > 0,
    ### )
    ### contained_terms = contained_term_query.all()

    return terms_matching_tokens + contained_terms


class RenderableSentence:
    """
    A collection of TextItems to be rendered.
    """

    def __init__(self, sentence_id, textitems):
        self.sentence_id = sentence_id
        self.textitems = textitems

    def __repr__(self):
        s = "".join([t.display_text for t in self.textitems])
        return f'<RendSent {self.sentence_id}, {len(self.textitems)} items, "{s}">'


def get_paragraphs(s, language):
    """
    Get array of arrays of RenderableSentences for the given string s.
    """
    terms = find_all_Terms_in_string(s, language)

    # Hacky reset of state of ParsedToken state.
    # _Shouldn't_ matter ... :-(
    ParsedToken.reset_counters()
    tokens = language.get_parsed_tokens(s)

    # Brutal hack ... for some reason the tests fail in
    # CI, but _inconsistently_, with the token order numbers.  The
    # order sometimes jumps by 2 ... I really can't explain it.  So,
    # as a _complete hack_, I'm re-numbering the tokens now, to ensure
    # they're in order.
    tokens.sort(key=lambda x: x.order)
    if len(tokens) > 0:
        n = tokens[0].order
        for t in tokens:
            t.order = n
            n += 1

    # Split into paragraphs.
    paragraphs = []
    curr_para = []
    for t in tokens:
        if t.token == "¶":
            paragraphs.append(curr_para)
            curr_para = []
        else:
            curr_para.append(t)
    if len(curr_para) > 0:
        paragraphs.append(curr_para)

    def make_RenderableSentence(pnum, sentence_num, tokens, terms):
        """
        Make a RenderableSentences using the tokens present in
        that sentence.  The current text and language are pulled
        into the function from the closure.
        """
        sentence_tokens = [t for t in tokens if t.sentence_number == sentence_num]
        renderable = RenderableCalculator.get_renderable(
            language, terms, sentence_tokens
        )
        textitems = [i.make_text_item(pnum, sentence_num, language) for i in renderable]
        ret = RenderableSentence(sentence_num, textitems)
        return ret

    def unique(arr):
        return list(set(arr))

    renderable_paragraphs = []
    pnum = 0
    for paratokens in paragraphs:
        # A renderable paragraph is a collection of RenderableSentences.
        renderable_sentences = [
            make_RenderableSentence(pnum, senum, paratokens, terms)
            for senum in sorted(unique([t.sentence_number for t in paratokens]))
        ]
        renderable_paragraphs.append(renderable_sentences)
        pnum += 1

    return renderable_paragraphs



"""
/read endpoints.
"""

from datetime import datetime
from flask import Blueprint, flash, request, render_template, redirect, jsonify
from lute.read.service import set_unknowns_to_known, start_reading
from lute.read.forms import TextForm
from lute.term.model import Repository
from lute.term.routes import handle_term_form
from lute.models.book import Book, Text
from lute.models.term import Term as DBTerm
from lute.models.setting import UserSetting
from lute.db import db


bp = Blueprint("read", __name__, url_prefix="/read")


def _render_book_page(book, pagenum):
    """
    Render a particular book page.
    """
    lang = book.language
    show_highlights = bool(int(UserSetting.get_value("show_highlights")))

    return render_template(
        "read/index.html",
        hide_top_menu=True,
        is_rtl=lang.right_to_left,
        html_title=book.title,
        book=book,
        dictionary_url=lang.sentence_translate_uri,
        page_num=pagenum,
        page_count=book.page_count,
        show_highlights=show_highlights,
    )


@bp.route("/<int:bookid>", methods=["GET"])
def read(bookid):
    """
    Read a book, opening to its current page.

    This is called from the book listing, on Lute index.
    """
    book = Book.find(bookid)
    if book is None:
        flash(f"No book matching id {bookid}")
        return redirect("/", 302)

    page_num = 1
    text = book.texts[0]
    if book.current_tx_id:
        text = Text.find(book.current_tx_id)
        page_num = text.order

    return _render_book_page(book, page_num)


@bp.route("/<int:bookid>/page/<int:pagenum>", methods=["GET"])
def read_page(bookid, pagenum):
    """
    Read a particular page of a book.

    Called from term Sentences link.
    """
    book = Book.find(bookid)
    if book is None:
        flash(f"No book matching id {bookid}")
        return redirect("/", 302)

    pagenum = book.page_in_range(pagenum)
    return _render_book_page(book, pagenum)


@bp.route("/page_done", methods=["post"])
def page_done():
    "Handle POST when page is done."
    data = request.json
    bookid = int(data.get("bookid"))
    pagenum = int(data.get("pagenum"))
    restknown = data.get("restknown")

    book = Book.find(bookid)
    text = book.text_at_page(pagenum)
    text.read_date = datetime.now()
    db.session.add(text)
    db.session.commit()
    if restknown:
        set_unknowns_to_known(text)
    return jsonify("ok")


@bp.route("/save_player_data", methods=["post"])
def save_player_data():
    "Save current player position, bookmarks.  Called on a loop by the player."
    data = request.json
    bookid = int(data.get("bookid"))
    book = Book.find(bookid)
    book.audio_current_pos = float(data.get("position"))
    book.audio_bookmarks = data.get("bookmarks")
    db.session.add(book)
    db.session.commit()
    return jsonify("ok")


@bp.route("/renderpage/<int:bookid>/<int:pagenum>", methods=["GET"])
def render_page(bookid, pagenum):
    "Method called by ajax, render the given page."
    book = Book.find(bookid)
    if book is None:
        flash(f"No book matching id {bookid}")
        return redirect("/", 302)
    paragraphs = start_reading(book, pagenum, db.session)
    return render_template("read/page_content.html", paragraphs=paragraphs)


@bp.route("/empty", methods=["GET"])
def empty():
    "Show an empty/blank page."
    return ""


@bp.route("/termform/<int:langid>/<text>", methods=["GET", "POST"])
def term_form(langid, text):
    """
    Create or edit a term.
    """
    repo = Repository(db)
    term = repo.find_or_new(langid, text)

    return handle_term_form(
        term,
        repo,
        "/read/frameform.html",
        render_template("/read/updated.html", term_text=term.text),
        embedded_in_reading_frame=True,
    )


@bp.route("/termpopup/<int:termid>", methods=["GET"])
def term_popup(termid):
    """
    Show a term popup for the given DBTerm.
    """
    term = DBTerm.query.get(termid)

    term_tags = [tt.text for tt in term.term_tags]

    def make_array(t):
        ret = {
            "term": t.text,
            "roman": t.romanization,
            "trans": t.translation if t.translation else "-",
            "tags": [tt.text for tt in t.term_tags],
        }
        return ret

    parent_terms = [p.text for p in term.parents]
    parent_terms = ", ".join(parent_terms)

    parent_data = []
    if len(term.parents) == 1:
        parent = term.parents[0]
        if parent.translation != term.translation:
            parent_data.append(make_array(parent))
    else:
        parent_data = [make_array(p) for p in term.parents]

    images = [term.get_current_image()] if term.get_current_image() else []
    for p in term.parents:
        if p.get_current_image():
            images.append(p.get_current_image())

    images = list(set(images))

    return render_template(
        "read/termpopup.html",
        term=term,
        flashmsg=term.get_flash_message(),
        term_tags=term_tags,
        term_images=images,
        parentdata=parent_data,
        parentterms=parent_terms,
    )


@bp.route("/flashcopied", methods=["GET"])
def flashcopied():
    return render_template("read/flashcopied.html")


@bp.route("/editpage/<int:bookid>/<int:pagenum>", methods=["GET", "POST"])
def edit_page(bookid, pagenum):
    "Edit the text on a page."
    book = Book.find(bookid)
    text = book.text_at_page(pagenum)
    if text is None:
        return redirect("/", 302)
    form = TextForm(obj=text)

    if form.validate_on_submit():
        form.populate_obj(text)
        db.session.add(text)
        db.session.commit()
        return redirect(f"/read/{book.id}", 302)

    return render_template("read/page_edit_form.html", hide_top_menu=True, form=form)



"""
User entry point.

Start lute running on given port, or 5000 if not set.

e.g.

python -m lute.main --port 5001
"""

import os
import argparse
import shutil
import logging
from waitress import serve
from lute.app_factory import create_app
from lute.config.app_config import AppConfig

logging.getLogger("waitress.queue").setLevel(logging.ERROR)
logging.getLogger("natto").setLevel(logging.CRITICAL)


def _print(s):
    """
    Print message to stdout.
    """
    if isinstance(s, str):
        s = s.split("\n")
    msg = "\n".join(["  " + lin.strip() for lin in s])
    print(msg, flush=True)


def _create_prod_config_if_needed():
    """
    If config.yml is missing, create one from prod config.
    """
    config_file = os.path.join(AppConfig.configdir(), "config.yml")
    if not os.path.exists(config_file):
        prod_conf = os.path.join(AppConfig.configdir(), "config.yml.prod")
        shutil.copy(prod_conf, config_file)
        _print(["", "Using new production config.", ""])


def _create_app(config_file_path=None):
    """
    Configure and init the app.

    Uses config file if set (throws if doesn't exist);
    otherwise, uses the prod config, creating a prod config
    if necessary.
    """
    _print(["", "Starting Lute:"])

    if config_file_path is not None:
        _print([f"Using specified config: {config_file_path}"])
    elif os.path.exists("config.yml"):
        _print(["Using config.yml found in root"])
        config_file_path = "config.yml"
    else:
        _print(["Using default config"])
        _create_prod_config_if_needed()
        config_file_path = AppConfig.default_config_filename()

    app_config = AppConfig(config_file_path)

    _print(["", "Initializing app."])
    app = create_app(config_file_path, output_func=_print)
    _print(f"data path: {app_config.datapath}")
    _print(f"database: {app_config.dbfilename}")
    if app_config.is_docker:
        _print("(Note these are container paths, not host paths.)")

    close_msg = """
    When you're finished reading, stop this process
    with Ctrl-C or your system equivalent.
    """
    if app_config.is_docker:
        close_msg = """
        When you're finished reading, stop this container
        with Ctrl-C, docker compose stop, or docker stop <containerid>
        as appropriate.
        """
    _print(close_msg)

    return app


def _start(args):
    "Configure and start the app."
    app = _create_app(args.config)

    _print(
        f"""
    Lute is running.  Open a web browser, and go to:

    http://localhost:{args.port}
    """
    )
    serve(app, host="0.0.0.0", port=args.port)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Start lute.")
    parser.add_argument(
        "--port", type=int, default=5000, help="Port number (default: 5000)"
    )
    parser.add_argument(
        "--config",
        help="Path to override config file.  Uses lute/config/config.yml if not set.",
    )
    try:
        _start(parser.parse_args())
    except Exception as e:  # pylint: disable=broad-exception-caught
        print("\n")
        print("-" * 50)
        print("Error during startup:")
        print(e)
        print("Please try again, or report an issue on GitHub.")
        print("-" * 50)



"""
Term mapping tests.
"""

import pytest
from lute.models.term import Term, TermTag, TermTextChangedException
from lute.db import db
from tests.dbasserts import assert_sql_result


def test_save_and_remove(empty_db, english):
    "Smoke test of mappings and db records."
    sql = "select WoText, WoTextLC, WoTokenCount from words"
    assert_sql_result(sql, [], "empty table")

    term = Term(english, "ABC")

    db.session.add(term)
    db.session.commit()
    assert_sql_result(sql, ["ABC; abc; 1"], "have term")

    db.session.delete(term)
    db.session.commit()
    assert_sql_result(sql, [], "no terms")


def test_term_with_parent_and_tags(empty_db, spanish):
    "Misc data check - parent and tags are saved."
    term = Term(spanish, "HOLA")
    parent = Term(spanish, "PARENT")
    term.add_parent(parent)

    tg = TermTag("tag")
    term.add_term_tag(tg)

    db.session.add(term)
    db.session.commit()

    sql = "select WoID, WoText, WoTextLC from words"
    expected = ["1; HOLA; hola", "2; PARENT; parent"]
    assert_sql_result(sql, expected, "sanity check on save")

    assert_sql_result("select WpWoID, WpParentWoID from wordparents", ["1; 2"], "wp")

    # Hacky sql check.
    sql = """select w.WoText, p.WoText as ptext, tags.TgText
        FROM words w
        INNER JOIN wordparents on WpWoID = w.WoID
        INNER JOIN words p on p.WoID = wordparents.WpParentWoID
        INNER JOIN wordtags on WtWoID = w.WoID
        INNER JOIN tags on TgID = WtTgID"""
    exp = ["HOLA; PARENT; tag"]
    assert_sql_result(sql, exp, "parents, tags")


def test_term_parent_with_two_children(spanish):
    "Both children get saved in the db."
    term = Term(spanish, "HOLA")
    gato = Term(spanish, "gato")
    parent = Term(spanish, "PARENT")
    gato.add_parent(parent)
    term.add_parent(parent)

    db.session.add(term)
    db.session.add(gato)
    db.session.commit()

    parent_get = Term.find(parent.id)
    assert parent_get.text == "PARENT"
    assert len(parent_get.children) == 2

    expected = [f"{term.id}; {parent.id}", f"{gato.id}; {parent.id}"]
    expected.sort()
    sql = """select WpWoID, WpParentWoID
    from wordparents order by WpWoID"""
    assert_sql_result(sql, expected, "wp")


def test_add_and_remove_parents(spanish):
    "Parent mappings."
    term = Term(spanish, "HOLA")
    p1 = Term(spanish, "parent")
    p2 = Term(spanish, "otro")
    term.add_parent(p1)
    term.add_parent(p2)

    db.session.add(term)
    db.session.commit()

    sql = """select WpWoID, WpParentWoID
    from wordparents order by WpWoID, WpParentWoID"""

    expected = [
        f"{term.id}; {p1.id}",
        f"{term.id}; {p2.id}",
    ]
    expected.sort()
    assert_sql_result(sql, expected, "after change")

    term.remove_all_parents()
    db.session.add(term)
    db.session.commit()
    assert_sql_result(sql, [], "all removed")


def test_remove_parent_leaves_children_in_db(spanish):
    "Child deletion doesn't delete the parent."
    term = Term(spanish, "HOLA")
    parent = Term(spanish, "PARENT")
    term.add_parent(parent)

    db.session.add(term)
    db.session.commit()
    sql_list = "select WoText from words order by WoText"
    sql = f"""select w.WoText, p.WoText as ptext from words w
    left join wordparents on WpWoID = w.WoID
    left join words p on p.WoID = wordparents.WpParentWoID
    where w.WoID = {term.id}"""
    assert_sql_result(sql_list, ["HOLA", "PARENT"], "both exist")
    assert_sql_result(sql, ["HOLA; PARENT"], "parent set")

    db.session.delete(parent)
    db.session.commit()
    assert_sql_result(sql_list, ["HOLA"], "parent removed")
    assert_sql_result(sql, ["HOLA; None"], "parent not set")


def test_removing_term_with_parent_and_tag(empty_db, spanish):
    """
    Both the parent and the tag are left.
    """
    term = Term(spanish, "HOLA")
    parent = Term(spanish, "PARENT")
    term.add_parent(parent)
    tg = TermTag("tag")
    term.add_term_tag(tg)

    db.session.add(term)
    db.session.commit()
    sqllist = "select WoText from words order by WoText"
    sqltags = "select TgText from tags"
    assert_sql_result(sqllist, ["HOLA", "PARENT"], "both exist")
    assert_sql_result(sqltags, ["tag"], "tag exists")

    db.session.delete(term)
    db.session.commit()
    assert_sql_result(sqllist, ["PARENT"], "parent left")
    assert_sql_result(sqltags, ["tag"], "tag left")


def test_save_replace_remove_image(spanish):
    "Save saves the associated image."
    t = Term(spanish, "HOLA")
    t.set_current_image("hello.png")
    assert t.get_current_image() == "hello.png"

    db.session.add(t)
    db.session.commit()
    sql = "select WiWoID, WiSource from wordimages"
    expected = ["1; hello.png"]
    assert_sql_result(sql, expected, "hello")

    t.set_current_image("there.png")
    assert t.get_current_image() == "there.png"

    db.session.add(t)
    db.session.commit()
    expected = ["1; there.png"]
    assert_sql_result(sql, expected, "there")

    t.set_current_image(None)
    assert t.get_current_image() is None

    db.session.add(t)
    db.session.commit()
    expected = []
    assert_sql_result(sql, expected, "removed")


def test_delete_term_deletes_image(spanish):
    "Check cascade delete."
    t = Term(spanish, "HOLA")
    t.set_current_image("hello.png")
    assert t.get_current_image() == "hello.png"

    db.session.add(t)
    db.session.commit()
    sql = "select WiWoID, WiSource from wordimages"
    expected = ["1; hello.png"]
    assert_sql_result(sql, expected, "hello")

    db.session.delete(t)
    db.session.commit()
    assert_sql_result(sql, [], "removed")


def test_save_remove_flash_message(spanish):
    "Flash message is associated with term, can be popped."
    t = Term(spanish, "HOLA")
    t.set_flash_message("hello")
    assert t.get_flash_message() == "hello"

    db.session.add(t)
    db.session.commit()
    sql = "select WfWoID, WfMessage from wordflashmessages"
    expected = ["1; hello"]
    assert_sql_result(sql, expected, "hello")

    assert t.pop_flash_message() == "hello", "popped"
    assert t.get_flash_message() is None, "no message now"

    db.session.add(t)
    db.session.commit()
    assert_sql_result(sql, [], "popped")


def test_delete_term_deletes_flash_message(spanish):
    "Check cascade delete."
    t = Term(spanish, "HOLA")
    t.set_flash_message("hello")

    db.session.add(t)
    db.session.commit()
    sql = "select WfWoID, WfMessage from wordflashmessages"
    expected = ["1; hello"]
    assert_sql_result(sql, expected, "hello")

    db.session.delete(t)
    db.session.commit()
    assert_sql_result(sql, [], "popped")


## Changing term text isn't allowed -- changing case is ok.


@pytest.mark.term_case
def test_changing_text_of_saved_Term_throws(english):
    "Changing text should throw."
    term = Term(english, "ABC")
    db.session.add(term)
    db.session.commit()

    with pytest.raises(TermTextChangedException):
        term.text = "DEF"


@pytest.mark.term_case
def test_changing_case_only_of_text_of_saved_Term_is_ok(english):
    "Changing text should throw."
    term = Term(english, "ABC")
    db.session.add(term)
    db.session.commit()

    term.text = "abc"


@pytest.mark.term_case
def test_changing_text_of_non_saved_Term_is_ok(english):
    "Changing text should throw."
    term = Term(english, "ABC")
    term.text = "DEF"



"""
DB setup tests using fake baseline, migration files.
"""

import os
import gzip
import sqlite3
from contextlib import closing
import pytest

from lute.db.setup.main import Setup, BackupManager
from lute.db.setup.migrator import SqliteMigrator


class MockBackupManager:  # pylint: disable=too-few-public-methods
    """
    Fake backup manager to use for tests.
    Can't figure out how to use monkeypatching or
    mocking for objects, so this stub will do
    for now.
    """

    def __init__(self):
        self.backup_called = False

    def do_backup(self):
        """
        Record if backup was called.
        """
        self.backup_called = True


class MockMigrator:  # pylint: disable=too-few-public-methods
    """
    Fake migrator to use for tests.
    """

    def __init__(self):
        self.has_pending_migrations = False
        self.migrations_run = False

    def has_migrations(self, conn):  # pylint: disable=unused-argument
        """
        Refs self.has_pending_migrations.
        """
        return self.has_pending_migrations

    def do_migration(self, conn):  # pylint: disable=unused-argument
        """
        Record if migrations run.
        """
        self.migrations_run = True


@pytest.fixture(name="fakebackupmanager")
def fixture_fakebackupmanager():
    """
    Fake object for tests.
    """
    return MockBackupManager()


@pytest.fixture(name="fakemigrator")
def fixture_fakemigrator():
    """
    Fake object for tests.
    """
    return MockMigrator()


@pytest.fixture(name="baseline")
def baseline_schema(tmp_path):
    """
    Baseline schema file used for setup.
    """
    baseline_file = tmp_path / "baseline.sql"
    content = "create table A (a int);"
    baseline_file.write_text(content)
    return baseline_file


def test_new_database(baseline, tmp_path, fakebackupmanager, fakemigrator):
    """
    If no db exists, setup should:
    - create the db using the baseline
    - run any migrations
    - no backup created
    """
    dbfile = tmp_path / "testdb.db"
    assert os.path.exists(dbfile) is False, "no db exists"

    fakemigrator.has_pending_migrations = True
    setup = Setup(dbfile, baseline, fakebackupmanager, fakemigrator)
    setup.setup()

    assert os.path.exists(dbfile), "db was created"
    assert fakemigrator.migrations_run, "migrations were run"
    assert fakebackupmanager.backup_called is False, "no backup called"


def test_existing_database_new_migrations(
    baseline, tmp_path, fakebackupmanager, fakemigrator
):
    """
    If db exists, setup should:
    - run any migrations
    - create a backup
    """
    dbfile = tmp_path / "testdb.db"
    dbfile.write_text("content")
    assert os.path.exists(dbfile), "db exists"

    fakemigrator.has_pending_migrations = True
    setup = Setup(dbfile, baseline, fakebackupmanager, fakemigrator)
    setup.setup()

    assert os.path.exists(dbfile), "db still there"
    assert fakemigrator.migrations_run, "migrations were run"
    assert fakebackupmanager.backup_called, "backup was called"


def test_existing_db_no_migrations(baseline, tmp_path, fakebackupmanager, fakemigrator):
    """
    DB should be left as-is, no migrations run, and *no* backup created.
    """
    dbfile = tmp_path / "testdb.db"
    dbfile.write_text("content")
    assert os.path.exists(dbfile), "db exists"

    fakemigrator.has_pending_migrations = False
    setup = Setup(dbfile, baseline, fakebackupmanager, fakemigrator)
    setup.setup()

    assert os.path.exists(dbfile), "db still there"
    assert fakemigrator.migrations_run, "migrations were run"
    assert fakebackupmanager.backup_called is False, "NO backup was called"


@pytest.fixture(name="migrator")
def fixture_migrator():
    """
    Migrator using the migration files in ./schema.
    """
    thisdir = os.path.dirname(os.path.realpath(__file__))
    migdir = os.path.join(thisdir, "schema", "migrations")
    repeatable = os.path.join(thisdir, "schema", "repeatable")
    return SqliteMigrator(migdir, repeatable)


def test_happy_path_no_existing_database(tmp_path, migrator):
    """
    If no db exists, setup should:
    - create the db using the baseline
    - run any migrations
    - no backup created
    """

    dbfile = tmp_path / "testdb.db"
    assert os.path.exists(dbfile) is False, "no db exists"

    backup_dir = tmp_path / "backups"
    backup_dir.mkdir()

    thisdir = os.path.dirname(os.path.realpath(__file__))
    baseline = os.path.join(thisdir, "schema", "baseline", "schema.sql")

    bm = BackupManager(dbfile, backup_dir, 3)

    setup = Setup(dbfile, baseline, bm, migrator)
    setup.setup()

    assert os.path.exists(dbfile), "db was created"

    with closing(sqlite3.connect(dbfile)) as conn:
        cur = conn.cursor()
        sql = """SELECT name FROM sqlite_master
        WHERE type='table' AND name in ('A', 'B')
        order by name;"""
        tnames = cur.execute(sql).fetchall()
        tnames = [t[0] for t in tnames]
        assert tnames == ["A", "B"], "migrations run"

    backup_files = list(backup_dir.glob("*.gz"))
    assert len(backup_files) == 0, "no backups"


def test_existing_database_with_migrations(tmp_path, migrator):
    """
    If db exists, setup should:
    - run any migrations
    - create a backup with today's datetime
    """

    dbfile = tmp_path / "testdb.db"
    assert os.path.exists(dbfile) is False, "no db exists"

    thisdir = os.path.dirname(os.path.realpath(__file__))
    baseline = os.path.join(thisdir, "schema", "baseline", "schema.sql")
    with open(baseline, "r", encoding="utf8") as f:
        with closing(sqlite3.connect(dbfile)) as conn:
            conn.executescript(f.read())
    assert os.path.exists(dbfile) is True, "db exists"

    def assert_tables(expected: list, msg: str):
        with closing(sqlite3.connect(dbfile)) as conn:
            cur = conn.cursor()
            sql = """SELECT name FROM sqlite_master
            WHERE type='table' AND name in ('A', 'B')
            order by name;"""
            tnames = cur.execute(sql).fetchall()
            tnames = [t[0] for t in tnames]
            assert tnames == expected, msg

    assert_tables(["A"], "initial state")

    backup_dir = tmp_path / "backups"
    backup_dir.mkdir()

    bm = BackupManager(dbfile, backup_dir, 3)

    setup = Setup(dbfile, baseline, bm, migrator)
    setup.setup()

    assert os.path.exists(dbfile), "db still exists"
    assert_tables(["A", "B"], "migrations run")

    backup_files = list(backup_dir.glob("*.gz"))
    print(backup_files)
    assert len(backup_files) == 1, "backup created"

    # Restore backup
    with gzip.open(backup_files[0], "rb") as gzipped_file, open(
        dbfile, "wb"
    ) as output_file:
        data = gzipped_file.read()
        output_file.write(data)

    assert_tables(["A"], "back to initial state")


def test_existing_database_no_outstanding_migrations(tmp_path, migrator):
    """
    If db exists and no migrations are outstand, setup should:
    - not run any migrations
    - NOT create a backup
    """

    dbfile = tmp_path / "testdb.db"
    assert os.path.exists(dbfile) is False, "no db exists"

    thisdir = os.path.dirname(os.path.realpath(__file__))
    baseline = os.path.join(thisdir, "schema", "baseline", "schema.sql")
    with open(baseline, "r", encoding="utf8") as f:
        with closing(sqlite3.connect(dbfile)) as conn:
            conn.executescript(f.read())
    assert os.path.exists(dbfile) is True, "db exists"

    backup_dir = tmp_path / "backups"
    backup_dir.mkdir()
    bm = BackupManager(dbfile, backup_dir, 3)

    setup = Setup(dbfile, baseline, bm, migrator)
    setup.setup()

    assert os.path.exists(dbfile), "db still exists"

    backup_files = list(backup_dir.glob("*.gz"))
    assert len(backup_files) == 1, "backup created"

    # Clean state before next run.
    for f in backup_files:
        os.unlink(f)
    backup_files = list(backup_dir.glob("*.gz"))
    assert len(backup_files) == 0, "no backups"

    setup.setup()
    backup_files = list(backup_dir.glob("*.gz"))
    assert len(backup_files) == 0, "STILL no backups"



"""
DataTables sqlite tests.
"""

from lute.utils.formutils import language_choices


def test_language_choices(app_context):
    "Gets all languages."
    choices = language_choices()
    assert choices[0][1] == "-", "- at the top"
    langnames = [c[1] for c in choices]
    assert "Spanish" in langnames, "have Spanish"



"""
App config tests
"""

import pytest
import yaml

from lute.config.app_config import AppConfig


def write_file(config_file, config_data):
    "Write config file."
    if "ENV" not in config_data:
        config_data["ENV"] = "dev"
    with open(config_file, "w", encoding="utf-8") as file:
        yaml.dump(config_data, file)


def test_valid_config(tmp_path):
    "Valid path and config is ok."
    config_file = tmp_path / "valid_config.yaml"
    config_data = {"DBNAME": "my_db", "DATAPATH": "data_path"}
    write_file(config_file, config_data)

    app_config = AppConfig(config_file)
    assert app_config.datapath == "data_path"
    assert app_config.sqliteconnstring == "sqlite:///data_path/my_db"
    assert app_config.env == "dev"


def test_ENV_required(tmp_path):
    "Throws otherwise."
    config_file = tmp_path / "valid_config.yaml"
    config_data = {"DBNAME": "my_db", "DATAPATH": "data_path"}
    with open(config_file, "w", encoding="utf-8") as file:
        yaml.dump(config_data, file)
    with pytest.raises(ValueError, match="ENV must be prod or dev, was None."):
        AppConfig(config_file)


def test_missing_dbname_throws(tmp_path):
    """
    File must contain DBNAME.  Testing/dev environment will have
    test_lute.db, prod will have an actual filename.
    """
    config_file = tmp_path / "missing_dbname.yaml"
    config_data = {"DATAPATH": "data_path"}
    write_file(config_file, config_data)

    with pytest.raises(ValueError, match="Config file must have 'DBNAME'"):
        AppConfig(config_file)


def test_system_specific_datapath_returned_if_DATAPATH_not_specified(tmp_path):
    """
    Using library to get platform-specific paths.  Tests will
    hardcode the appropriate system path.
    """
    config_file = tmp_path / "default_datapath.yaml"
    config_data = {"DBNAME": "my_db"}
    write_file(config_file, config_data)

    app_config = AppConfig(config_file)

    # Can't test the exact path returned, it will vary by system.
    # Could monkeypatch/dependency inject, but too much hassle.
    assert "Lute3" in app_config.datapath


def test_env_can_only_be_prod_or_dev(tmp_path):
    """
    Throws error if not prod or dev.
    """
    config_file = tmp_path / "default_datapath.yaml"
    config_data = {"DBNAME": "my_db", "ENV": "blah"}
    write_file(config_file, config_data)

    with pytest.raises(ValueError, match="ENV must be prod or dev, was blah."):
        AppConfig(config_file)


def test_invalid_yaml_throws(tmp_path):
    "File must be valid."
    config_file = tmp_path / "invalid_yaml.yaml"
    with open(config_file, "w", encoding="utf-8") as f:
        f.write("bad_data")

    with pytest.raises(Exception):
        AppConfig(config_file)


def test_nonexistent_config_file_throws(tmp_path):
    """
    File must exist!  During packaging, a file will be generated
    in the config folder.
    """
    config_file = tmp_path / "nonexistent_config.yaml"
    with pytest.raises(FileNotFoundError, match="No such file"):
        AppConfig(config_file)



"""
/language endpoints.
"""

from sqlalchemy import func
from sqlalchemy.exc import IntegrityError
from flask import Blueprint, current_app, render_template, redirect, url_for, flash
from lute.models.language import Language
from lute.models.book import Book
from lute.models.term import Term
from lute.language.forms import LanguageForm
from lute.db import db
from lute.db.demo import predefined_languages
from lute.parse.registry import supported_parsers

bp = Blueprint("language", __name__, url_prefix="/language")


@bp.route("/index")
def index():
    """
    List all languages.

    This includes the Book and Term count for each Language.  These
    counts are pulled in by subqueries, because Language doesn't have
    "books" and "terms" members ... I was having trouble with session
    management when these were added, and they're only used here, so
    this is good enough for now.
    """

    def create_count_subquery(class_, count_column):
        # Re the pylint disable, ref
        # https://github.com/pylint-dev/pylint/issues/8138 ...
        ret = (
            db.session.query(
                class_.language_id,
                # pylint: disable=not-callable
                func.count(class_.id).label(count_column),
            )
            .group_by(class_.language_id)
            .subquery()
        )
        return ret

    # Create subqueries for counting books and terms
    book_subquery = create_count_subquery(Book, "book_count")
    term_subquery = create_count_subquery(Term, "term_count")

    # Query to join Language with book and term counts
    query = (
        db.session.query(
            Language, book_subquery.c.book_count, term_subquery.c.term_count
        )
        .outerjoin(book_subquery, Language.id == book_subquery.c.language_id)
        .outerjoin(term_subquery, Language.id == term_subquery.c.language_id)
    )

    results = query.all()

    results = [rec for rec in results if rec[0].is_supported is True]

    return render_template("language/index.html", language_data=results)


def _handle_form(language, form) -> bool:
    """
    Handle the language form processing.

    Returns True if validated and saved.
    """
    result = False

    if form.validate_on_submit():
        try:
            form.populate_obj(language)
            current_app.db.session.add(language)
            current_app.db.session.commit()
            flash(f"Language {language.name} updated", "success")
            result = True
        except IntegrityError as e:
            msg = e.orig
            if "languages.LgName" in f"{e.orig}":
                msg = f"{language.name} already exists."
            flash(msg, "error")

    return result


@bp.route("/edit/<int:langid>", methods=["GET", "POST"])
def edit(langid):
    """
    Edit a language.
    """
    language = db.session.get(Language, langid)

    if not language:
        flash(f"Language {langid} not found", "danger")
        return redirect(url_for("language.index"))

    form = LanguageForm(obj=language)
    form.parser_type.choices = supported_parsers()

    if _handle_form(language, form):
        return redirect("/")
    return render_template("language/edit.html", form=form, language=language)


@bp.route("/new", defaults={"langname": None}, methods=["GET", "POST"])
@bp.route("/new/<string:langname>", methods=["GET", "POST"])
def new(langname):
    """
    Create a new language.
    """
    predefined = predefined_languages()
    language = Language()
    if langname is not None:
        candidates = [lang for lang in predefined if lang.name == langname]
        if len(candidates) == 1:
            language = candidates[0]

    form = LanguageForm(obj=language)
    form.parser_type.choices = supported_parsers()

    if _handle_form(language, form):
        return redirect("/")

    return render_template(
        "language/new.html", form=form, language=language, predefined=predefined
    )


@bp.route("/delete/<int:langid>", methods=["POST"])
def delete(langid):
    """
    Delete a language.
    """
    language = db.session.get(Language, langid)
    if not language:
        flash(f"Language {langid} not found")
    Language.delete(language)
    return redirect(url_for("language.index"))



"""
Book entity.
"""

from lute.db import db
from lute.parse.base import SentenceGroupIterator

booktags = db.Table(
    "booktags",
    db.Model.metadata,
    db.Column("BtT2ID", db.Integer, db.ForeignKey("tags2.T2ID")),
    db.Column("BtBkID", db.Integer, db.ForeignKey("books.BkID")),
)


class BookTag(db.Model):
    "Term tags."
    __tablename__ = "tags2"

    id = db.Column("T2ID", db.Integer, primary_key=True)
    text = db.Column("T2Text", db.String(20))
    comment = db.Column("T2Comment", db.String(200))

    @staticmethod
    def make_book_tag(text, comment=""):
        "Create a BookTag."
        tt = BookTag()
        tt.text = text
        tt.comment = comment
        return tt

    @staticmethod
    def find_by_text(text):
        "Find a tag by text, or None if not found."
        return db.session.query(BookTag).filter(BookTag.text == text).first()

    @staticmethod
    def find_or_create_by_text(text):
        "Return tag or create one."
        ret = BookTag.find_by_text(text)
        if ret is not None:
            return ret
        return BookTag.make_book_tag(text)


class Book(
    db.Model
):  # pylint: disable=too-few-public-methods, too-many-instance-attributes
    """
    Book entity.
    """

    __tablename__ = "books"

    id = db.Column("BkID", db.SmallInteger, primary_key=True)
    title = db.Column("BkTitle", db.String(length=200))
    language_id = db.Column(
        "BkLgID", db.Integer, db.ForeignKey("languages.LgID"), nullable=False
    )
    source_uri = db.Column("BkSourceURI", db.String(length=1000))
    current_tx_id = db.Column("BkCurrentTxID", db.Integer, default=0)
    archived = db.Column("BkArchived", db.Boolean, default=False)

    audio_filename = db.Column("BkAudioFilename", db.String)
    audio_current_pos = db.Column("BkAudioCurrentPos", db.Float)
    audio_bookmarks = db.Column("BkAudioBookmarks", db.String)

    language = db.relationship("Language")
    texts = db.relationship(
        "Text",
        back_populates="book",
        order_by="Text.order",
        cascade="all, delete-orphan",
    )
    book_tags = db.relationship("BookTag", secondary="booktags")

    def __init__(self, title=None, language=None, source_uri=None):
        self.title = title
        self.language = language
        self.source_uri = source_uri
        self.texts = []
        self.book_tags = []

    def __repr__(self):
        return f"<Book {self.id} {self.title}>"

    def remove_all_book_tags(self):
        self.book_tags = []

    def add_book_tag(self, book_tag):
        if book_tag not in self.book_tags:
            self.book_tags.append(book_tag)

    def remove_book_tag(self, book_tag):
        self.book_tags.remove(book_tag)

    @property
    def page_count(self):
        return len(self.texts)

    def page_in_range(self, n):
        "Return page number that is in the book's page count."
        ret = max(n, 1)
        ret = min(ret, self.page_count)
        return ret

    def text_at_page(self, n):
        "Return the text object at page n."
        pagenum = self.page_in_range(n)
        return self.texts[pagenum - 1]

    @property
    def is_supported(self):
        "True if the book's language's parser is supported."
        return self.language.is_supported

    @staticmethod
    def create_book(title, language, fulltext, max_word_tokens_per_text=250):
        """
        Create a book with given fulltext content,
        splitting the content into separate Text objects with max
        token count.
        """
        tokens = language.parser.get_parsed_tokens(fulltext, language)

        def token_string(toks):
            a = [t.token for t in toks]
            ret = "".join(a)
            ret = ret.replace("\r", "")
            ret = ret.replace("¶", "\n")
            return ret.strip()

        b = Book(title, language)
        page_number = 0
        it = SentenceGroupIterator(tokens, max_word_tokens_per_text)
        while toks := it.next():
            page_number += 1
            # Note the text is automatically added to b.texts!
            t = Text(b, token_string(toks), page_number)

        return b

    @staticmethod
    def find(book_id):
        "Get by ID."
        return db.session.query(Book).filter(Book.id == book_id).first()


# TODO zzfuture fix: rename class and table to Page/pages
class Text(db.Model):
    """
    Each page in a Book.
    """

    __tablename__ = "texts"

    id = db.Column("TxID", db.Integer, primary_key=True)
    _text = db.Column("TxText", db.String, nullable=False)
    order = db.Column("TxOrder", db.Integer)
    _read_date = db.Column("TxReadDate", db.DateTime, nullable=True)
    bk_id = db.Column("TxBkID", db.Integer, db.ForeignKey("books.BkID"), nullable=False)
    word_count = db.Column("TxWordCount", db.Integer, nullable=True)

    book = db.relationship("Book", back_populates="texts")
    sentences = db.relationship(
        "Sentence",
        back_populates="text",
        order_by="Sentence.order",
        cascade="all, delete-orphan",
    )

    def __init__(self, book, text, order=1):
        self.book = book
        self.text = text
        self.order = order
        self.sentences = []

    @property
    def title(self):
        """
        Text title is the book title + page fraction.
        """
        b = self.book
        s = f"({self.order}/{self.book.page_count})"
        t = f"{b.title} {s}"
        return t

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, s):
        self._text = s
        toks = self._get_parsed_tokens()
        wordtoks = [t for t in toks if t.is_word]
        self.word_count = len(wordtoks)
        if self._read_date is not None:
            self._load_sentences_from_tokens(toks)

    @property
    def read_date(self):
        return self._read_date

    @read_date.setter
    def read_date(self, s):
        self._read_date = s
        # Ensure loaded.
        self.load_sentences()

    def _get_parsed_tokens(self):
        "Return the tokens."
        lang = self.book.language
        return lang.parser.get_parsed_tokens(self.text, lang)

    def _load_sentences_from_tokens(self, parsedtokens):
        "Save sentences using the tokens."
        self._remove_sentences()
        curr_sentence_tokens = []
        sentence_number = 1
        for pt in parsedtokens:
            curr_sentence_tokens.append(pt)
            if pt.is_end_of_sentence:
                se = Sentence.from_tokens(curr_sentence_tokens, sentence_number)
                self._add_sentence(se)
                # Reset for the next sentence.
                curr_sentence_tokens = []
                sentence_number += 1

        # Add any stragglers.
        if len(curr_sentence_tokens) > 0:
            se = Sentence.from_tokens(curr_sentence_tokens, sentence_number)
            self._add_sentence(se)

    def load_sentences(self):
        """
        Parse the current text and create Sentence objects.
        """
        toks = self._get_parsed_tokens()
        self._load_sentences_from_tokens(toks)

    def _add_sentence(self, sentence):
        "Add a sentence to the Text."
        if sentence not in self.sentences:
            self.sentences.append(sentence)
            sentence.text = self

    def _remove_sentences(self):
        "Remove all sentence from the Text."
        for sentence in self.sentences:
            sentence.text = None
        self.sentences = []

    @staticmethod
    def find(text_id):
        "Get by ID."
        return db.session.query(Text).filter(Text.id == text_id).first()


class Sentence(db.Model):
    """
    Parsed sentences for a given Text.

    The Sentence contains the parsed tokens, joined by the zero-width string.
    """

    __tablename__ = "sentences"

    id = db.Column("SeID", db.Integer, primary_key=True)
    tx_id = db.Column("SeTxID", db.Integer, db.ForeignKey("texts.TxID"), nullable=False)
    order = db.Column("SeOrder", db.Integer, default=1)
    text_content = db.Column("SeText", db.Text, default="")

    text = db.relationship("Text", back_populates="sentences")

    def __init__(self, text_content="", text=None, order=1):
        self.text_content = text_content
        self.text = text
        self.order = order

    @staticmethod
    def from_tokens(tokens, senumber):
        """
        Create a new Sentence from ParsedTokens.
        """

        ptstrings = [t.token for t in tokens]

        zws = chr(0x200B)  # Zero-width space.
        s = zws.join(ptstrings)
        s = s.strip(" ")

        # The zws is added at the start and end of each
        # sentence, to standardize the string search when
        # looking for terms.
        s = zws + s + zws

        sentence = Sentence()
        sentence.order = senumber
        sentence.text_content = s
        return sentence



"""
Functions to manage demo database.

Lute db comes pre-loaded with some demo data.  User can view Tutorial,
wipe data, etc.

The db settings table contains a record, StKey = 'IsDemoData', if the
data is demo.
"""

import os
import re
from glob import glob
import yaml
from sqlalchemy import text

from lute.models.language import Language
from lute.models.book import Book
from lute.book.stats import refresh_stats
from lute.models.setting import SystemSetting
from lute.db import db
import lute.db.management


def contains_demo_data():
    """
    True if IsDemoData setting is present.
    """
    ss = SystemSetting.get_value("IsDemoData")
    if ss is None:
        return False
    return True


def remove_flag():
    """
    Remove IsDemoData setting.
    """
    if not contains_demo_data():
        raise RuntimeError("Can't delete non-demo data.")

    SystemSetting.delete_key("IsDemoData")
    db.session.commit()


def tutorial_book_id():
    """
    Return the book id of the tutorial.
    """
    if not contains_demo_data():
        return None
    sql = """select BkID from books
    inner join languages on LgID = BkLgID
    where LgName = 'English' and BkTitle = 'Tutorial'
    """
    r = db.session.execute(text(sql)).first()
    if r is None:
        return None
    return int(r[0])


def delete_demo_data():
    """
    If this is a demo, wipe everything.
    """
    if not contains_demo_data():
        raise RuntimeError("Can't delete non-demo data.")
    remove_flag()
    lute.db.management.delete_all_data()


# Loading demo data.


def demo_data_path():
    """
    Path to the demo data yaml files.
    """
    thisdir = os.path.dirname(__file__)
    demo_dir = os.path.join(thisdir, "demo")
    return os.path.abspath(demo_dir)


def get_demo_language(filename):
    """
    Create a new Language object from a yaml definition.
    """
    with open(filename, "r", encoding="utf-8") as file:
        d = yaml.safe_load(file)

    lang = Language()

    def load(key, method):
        if key in d:
            val = d[key]
            # Handle boolean values
            if isinstance(val, str):
                temp = val.lower()
                if temp == "true":
                    val = True
                elif temp == "false":
                    val = False
            setattr(lang, method, val)

    # Define mappings for fields
    mappings = {
        "name": "name",
        "dict_1": "dict_1_uri",
        "dict_2": "dict_2_uri",
        "sentence_translation": "sentence_translate_uri",
        "show_romanization": "show_romanization",
        "right_to_left": "right_to_left",
        "parser_type": "parser_type",
        "character_substitutions": "character_substitutions",
        "split_sentences": "regexp_split_sentences",
        "split_sentence_exceptions": "exceptions_split_sentences",
        "word_chars": "word_characters",
    }

    for key in d.keys():
        funcname = mappings.get(key, "")
        if funcname:
            load(key, funcname)

    return lang


def predefined_languages():
    "Languages that have yaml files."
    demo_glob = os.path.join(demo_data_path(), "languages", "*.yaml")
    langs = [get_demo_language(f) for f in glob(demo_glob)]
    langs.sort(key=lambda x: x.name)
    return langs


def load_demo_languages():
    """
    Load predefined languages.  Assume everything is supported.

    This method will also be called during acceptance tests, so it's "public".
    """
    supported = [lang for lang in predefined_languages() if lang.is_supported]
    for lang in supported:
        db.session.add(lang)
    db.session.commit()


def load_demo_stories():
    "Load the stories."
    demo_glob = os.path.join(demo_data_path(), "stories", "*.txt")
    for filename in glob(demo_glob):
        with open(filename, "r", encoding="utf-8") as f:
            content = f.read()

        langpatt = r"language:\s*(.*)\n"
        lang = re.search(langpatt, content).group(1).strip()
        lang = Language.find_by_name(lang)

        if lang is None or not lang.is_supported:
            pass
        else:
            title_match = re.search(r"title:\s*(.*)\n", content)
            title = title_match.group(1).strip()
            content = re.sub(r"#.*\n", "", content)
            b = Book.create_book(title, lang, content)
            db.session.add(b)
    SystemSetting.set_value("IsDemoData", True)
    db.session.commit()
    refresh_stats()


def load_demo_data():
    """
    Load the data.
    """
    load_demo_languages()
    load_demo_stories()
    SystemSetting.set_value("IsDemoData", True)
    db.session.commit()



"""
Common form methods.
"""

from lute.models.language import Language
from lute.db import db


def language_choices():
    """
    Return the list of languages for select boxes.

    If only one lang exists, only return that,
    otherwise add a '-' dummy entry at the top.
    """
    langs = db.session.query(Language).order_by(Language.name).all()
    supported = [lang for lang in langs if lang.is_supported]
    lang_choices = [(s.id, s.name) for s in supported]
    if len(lang_choices) > 1:
        lang_choices = [(0, "-")] + lang_choices
    return lang_choices
